import os
import ast
import sys
import asyncio
from pathlib import Path
from runpy import run_path
from pydantic._internal._model_construction import ModelMetaclass

from panther import status
from panther.request import Request
from panther.response import Response
from panther.exceptions import APIException
from panther.configs import JWTConfig, config
from panther.middlewares.base import BaseMiddleware
from panther.middlewares.monitoring import Middleware as MonitoringMiddleware
from panther.routings import find_endpoint, check_and_load_urls, finalize_urls, flatten_urls, collect_path_variables
from panther._utils import http_response, import_class, read_body

""" We can't import logger on the top cause it needs config['base_dir'] ans its fill in __init__ """


class Panther:

    def __init__(self, name):
        from panther.logger import logger
        os.system('clear')
        config['base_dir'] = Path(name).resolve().parent
        self.panther_dir = Path(__file__).parent
        if sys.version_info.minor < 11:
            logger.warning('Use Python Version 3.11+ For Better Performance.')
        self.load_configs()

    def load_configs(self) -> None:
        from panther.logger import logger
        logger.debug(f'Base directory: {config["base_dir"]}')

        # Check & Read The Configs File
        self._check_configs()

        # Put Variables In "config"
        config['monitoring'] = self.settings.get('MONITORING', config['monitoring'])
        config['log_queries'] = self.settings.get('LOG_QUERIES', config['log_queries'])
        config['default_cache_exp'] = self.settings.get('DEFAULT_CACHE_EXP', config['default_cache_exp'])
        config['throttling'] = self.settings.get('THROTTLING', config['throttling'])
        config['secret_key'] = self._get_secret_key()

        config['middlewares'] = self._get_middlewares()
        config['reversed_middlewares'] = config['middlewares'][::-1]
        config['user_model'] = self._get_user_model()

        config['authentication'] = self._get_authentication_class()
        config['jwt_config'] = self._get_jwt_config()

        # Find Database Models
        self._collect_models()

        # Check & Collect URLs
        #   check_urls should be the last call in load_configs,
        #   because it will read all files and load them.
        config['urls'] = self._load_urls()

        # This import shouldn't be on top
        from panther.panel.urls import urls as panel_urls
        config['urls']['_panel'] = panel_urls

        logger.debug('Configs loaded.')
        if config['monitoring']:
            logger.info('Run "panther monitor" in another session for Monitoring.')

    def _check_configs(self):
        from panther.logger import logger
        """Read the config file and put it as dict in self.settings"""
        try:
            configs_path = config['base_dir'] / 'core/configs.py'
            self.settings = run_path(str(configs_path))
        except FileNotFoundError:
            logger.critical('core/configs.py Not Found.')
            # TODO: Exit() Here

    def _get_secret_key(self) -> bytes | None:
        if secret_key := self.settings.get('SECRET_KEY'):
            return secret_key.encode()
        return secret_key

    def _get_middlewares(self) -> list:
        """Collect The Middlewares & Set db_engine If One Of Middlewares Was For DB"""
        from panther.logger import logger
        middlewares = list()

        for path, data in self.settings.get('MIDDLEWARES', []):
            if path.find('panther.middlewares.db.Middleware') != -1:
                config['db_engine'] = data['url'].split(':')[0]

            Middleware = import_class(path)  # NOQA: Py Pep8 Naming
            if not issubclass(Middleware, BaseMiddleware):
                logger.critical(f'{Middleware} is not a sub class of BaseMiddleware.')
                continue

            middlewares.append(Middleware(**data))  # NOQA: Py Argument List
        return middlewares

    def _get_user_model(self) -> ModelMetaclass:
        return import_class(self.settings.get('USER_MODEL', 'panther.db.models.BaseUser'))

    def _get_authentication_class(self) -> ModelMetaclass | None:
        return self.settings.get('AUTHENTICATION') and import_class(self.settings['AUTHENTICATION'])

    def _get_jwt_config(self) -> JWTConfig:
        """Only Collect JWT Config If Authentication Is JWTAuthentication"""
        if getattr(config['authentication'], '__name__', None) == 'JWTAuthentication':
            user_config = self.settings.get('JWTConfig')
            return JWTConfig(**user_config) if user_config else JWTConfig(key=config['secret_key'].decode())

    @classmethod
    def _collect_models(cls):
        """Collecting models for panel APIs"""
        from panther.db.models import Model

        for root, _, files in os.walk(config['base_dir']):
            # Traverse through each directory
            for f in files:
                # Traverse through each file of directory
                if f == 'models.py':
                    # If the file was "models.py" read it
                    file_path = f'{root}/models.py'
                    with open(file_path, 'r') as file:
                        # Parse the file with ast
                        node = ast.parse(file.read())
                        for n in node.body:
                            # Find classes in each element of files' body
                            if type(n) is ast.ClassDef and n.bases:
                                class_path = file_path\
                                    .removesuffix('/models.py')\
                                    .removeprefix(f'{config["base_dir"]}/')\
                                    .replace('/', '.')
                                # We don't need to import the package classes
                                if class_path.find('site-packages') == -1:
                                    # Import the class to check his father and brother
                                    klass = import_class(f'{class_path}.models.{n.name}')
                                    for parent in klass.__mro__:
                                        if parent is Model:
                                            # The class was one our database models so collect it
                                            config['models'].append({
                                                'name': n.name,
                                                'path':  file_path,
                                                'class': klass,
                                                'app': class_path.split('.'),
                                            })

    def _load_urls(self) -> dict:
        urls = check_and_load_urls(self.settings.get('URLs')) or {}
        collected_urls = flatten_urls(urls)
        return finalize_urls(collected_urls)

    async def __call__(self, scope, receive, send) -> None:
        """
        We Used Python3.11+ For asyncio.TaskGroup()
        1.
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.run(scope, receive, send))
        2.
            await self.run(scope, receive, send)
        3.
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(self.run, scope, receive, send)
                await anyio.to_thread.run_sync(self.run, scope, receive, send)
        4.
            with ProcessPoolExecutor() as e:
                e.submit(self.run, scope, receive, send)
        """
        if sys.version_info.minor >= 11:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.run(scope, receive, send))
        else:
            await self.run(scope, receive, send)

    async def run(self, scope, receive, send):
        from panther.logger import logger
        # Read Body & Create Request
        body = await read_body(receive)
        request = Request(scope=scope, body=body)

        # Monitoring Middleware
        monitoring_middleware = None
        if config['monitoring']:
            monitoring_middleware = MonitoringMiddleware()
            await monitoring_middleware.before(request=request)

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=request.path)
        path_variables = collect_path_variables(request_path=request.path, found_path=found_path)

        if endpoint is None:
            return await http_response(
                send, status_code=status.HTTP_404_NOT_FOUND, monitoring=monitoring_middleware, exception=True,
            )

        try:  # They Both Have The Save Exception (APIException)
            # Call 'Before' Middlewares
            for middleware in config['middlewares']:
                request = await middleware.before(request=request)

            # User Didn't Use @API Decorator
            if not hasattr(endpoint, '__wrapped__'):
                logger.critical(f'You may have forgotten to use @API on the {endpoint.__name__}()')
                return await http_response(
                    send,
                    status_code=status.HTTP_510_NOT_EXTENDED,
                    monitoring=monitoring_middleware,
                    exception=True,
                )

            # Call Endpoint
            response = await endpoint(request=request, **path_variables)
        except APIException as e:
            response = self.handle_exceptions(e)
        except Exception as e:
            # Every unhandled exception in Panther or code will catch here
            logger.critical(e)
            return await http_response(
                send,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                monitoring=monitoring_middleware,
                exception=True,
            )

        # Call 'After' Middleware
        for middleware in config['reversed_middlewares']:
            try:
                response = await middleware.after(response=response)
            except APIException as e:
                response = self.handle_exceptions(e)

        await http_response(
            send, status_code=response.status_code, monitoring=monitoring_middleware, body=response.body,
        )

    @classmethod
    def handle_exceptions(cls, e, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )



import random
from unittest import TestCase

from panther.routings import flatten_urls, finalize_urls, find_endpoint, collect_path_variables


class TestRoutingFunctions(TestCase):

    def test_check_and_load_urls(self):
        # TODO: ...
        pass

    def test_collect_ellipsis_urls(self):
        urls = {
            'user/': {
                '<user_id>/': ...,
                'profile/': ...,
                'list/': ...,
            },
        }

        with self.assertLogs() as captured:
            collected_urls = flatten_urls(urls)

        self.assertEqual(len(captured.records), 3)
        self.assertEqual(captured.records[0].getMessage(), "URL Can't Point To Ellipsis. ('user/<user_id>/' -> ...)")
        self.assertEqual(captured.records[1].getMessage(), "URL Can't Point To Ellipsis. ('user/profile/' -> ...)")
        self.assertEqual(captured.records[2].getMessage(), "URL Can't Point To Ellipsis. ('user/list/' -> ...)")

        self.assertDictEqual(collected_urls, {})

    def test_collect_None_urls(self):
        urls = {
            'user/': {
                '<user_id>/': None,
                'profile/': None,
                'list/': None,
            },
        }

        with self.assertLogs() as captured:
            collected_urls = flatten_urls(urls)

        self.assertEqual(len(captured.records), 3)
        self.assertEqual(captured.records[0].getMessage(), "URL Can't Point To None. ('user/<user_id>/' -> None)")
        self.assertEqual(captured.records[1].getMessage(), "URL Can't Point To None. ('user/profile/' -> None)")
        self.assertEqual(captured.records[2].getMessage(), "URL Can't Point To None. ('user/list/' -> None)")

        self.assertDictEqual(collected_urls, {})

    def test_collect_invalid_urls(self):
        def temp_func(): pass

        urls = {
            'user/': {
                '?': temp_func,
                '%^': temp_func,
                'لیست': temp_func,
            },
        }

        with self.assertLogs() as captured:
            collected_urls = flatten_urls(urls)

        self.assertEqual(len(captured.records), 3)
        self.assertEqual(captured.records[0].getMessage(), "URL Is Not Valid. --> 'user/?/'")
        self.assertEqual(captured.records[1].getMessage(), "URL Is Not Valid. --> 'user/%^/'")
        self.assertEqual(captured.records[2].getMessage(), "URL Is Not Valid. --> 'user/لیست/'")

        self.assertDictEqual(collected_urls, {})

    def test_collect_empty_url(self):
        def temp_func(): pass

        urls = {
            '': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '': temp_func,
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_empty_url_nested(self):
        def temp_func(): pass

        urls = {
            '': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func,
            'nested/<user_id>/': temp_func,
            'nested/profile/': temp_func,
            'nested/list/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_root_url(self):
        def temp_func(): pass

        urls = {
            '/': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '/': temp_func,
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_root_url_nested(self):
        def temp_func(): pass

        urls = {
            '/': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '/<user_id>/': temp_func,
            '/profile/': temp_func,
            '/list/': temp_func,
            '//nested/<user_id>/': temp_func,
            '//nested/profile/': temp_func,
            '//nested/list/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_simple_urls(self):
        def temp_func(): pass

        urls = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_simple_nested_urls(self):
        def temp_func(): pass

        urls = {
            'user/': {
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            }
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            'user/<user_id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_simple_nested_urls_without_slash_at_end(self):
        def temp_func(): pass

        urls = {
            'user': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            }
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            'user/<user_id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_collect_complex_nested_urls(self):
        def temp_func(): pass

        urls = {
            'user/': {
                '<user_id>/profile/<id>': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
            '': {
                'payments': temp_func,
                'notifications': temp_func,
            },
            'admin/v1': {
                'profile/avatar': temp_func,
                '<user_id>': temp_func,
                'users/': {
                    'list': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                    'detail': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                }
            },
            'admin/v2': {

            }
        }

        collected_urls = flatten_urls(urls)
        expected_result = {
            'user/<user_id>/profile/<id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func,
            'payments/': temp_func,
            'notifications/': temp_func,
            'admin/v1/profile/avatar/': temp_func,
            'admin/v1/<user_id>/': temp_func,
            'admin/v1/users/list/registered/': temp_func,
            'admin/v1/users/list/not-registered/': temp_func,
            'admin/v1/users/detail/registered/': temp_func,
            'admin/v1/users/detail/not-registered/': temp_func
        }
        self.assertEqual(collected_urls, expected_result)

    def test_finalize_empty_url(self):
        def temp_func(): pass

        urls = {
            '': temp_func,
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            '': temp_func,
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_empty_url_nested(self):
        def temp_func(): pass

        urls = {
            '': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func
            },
            '<user_id>': temp_func,
            'profile': temp_func,
            'list': temp_func
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_root_url(self):
        def temp_func(): pass

        urls = {
            '/': temp_func,
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            '': temp_func,
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_root_url_nested(self):
        def temp_func(): pass

        urls = {
            '/': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func
            },
            '<user_id>': temp_func,
            'profile': temp_func,
            'list': temp_func
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_root_and_empty_url_nested(self):
        def temp_func(): pass

        urls = {
            '/': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func
            },
            '': temp_func,
            'profile': temp_func,
            'list': temp_func
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_empty_and_root_url_nested(self):
        def temp_func(): pass

        urls = {
            '': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func
            },
            '': temp_func,
            'profile': temp_func,
            'list': temp_func,
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_finalize_urls(self):
        def temp_func(): pass

        urls = {
            'user/': {
                '<user_id>/profile/<id>': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
            '': {
                'payments': temp_func,
                'notifications': temp_func,
            },
            'admin/v1': {
                'profile/avatar': temp_func,
                '<user_id>': temp_func,
                'users/': {
                    'list': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                    'detail': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                }
            },
            'admin/v2': {

            }
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': temp_func
                    }
                },
                'profile': temp_func,
                'list': temp_func
            },
            'payments': temp_func,
            'notifications': temp_func,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': temp_func
                    },
                    '<user_id>': temp_func,
                    'users': {
                        'list': {
                            'registered': temp_func,
                            'not-registered': temp_func
                        },
                        'detail': {
                            'registered': temp_func,
                            'not-registered': temp_func
                        }
                    }
                }
            }
        }
        self.assertEqual(finalized_urls, expected_result)

    def test_find_endpoint_root_url(self):
        def temp_func(): pass

        from panther.configs import config

        config['urls'] = {
            '': temp_func
        }
        _func, _ = find_endpoint('')

        self.assertIsNotNone(_func)
        self.assertEqual(_func, temp_func)
        config['urls'] = {}

    def test_find_endpoint_success(self):
        def user_id_profile_id(): pass

        def user_profile(): pass

        def payment(): pass

        def admin_v1_profile_avatar(): pass

        def admin_v1_id(): pass

        def admin_v2_users_list_registered(): pass

        def admin_v2_users_detail_not_registered(): pass

        from panther.configs import config

        config['urls'] = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': user_id_profile_id
                    }
                },
                'profile': user_profile,
                'list': ...
            },
            'payments': payment,
            'notifications': ...,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': admin_v1_profile_avatar
                    },
                    '<user_id>': admin_v1_id,
                    'users': {
                        'list': {
                            'registered': admin_v2_users_list_registered,
                            'not-registered': ...
                        },
                        'detail': {
                            'registered': ...,
                            'not-registered': admin_v2_users_detail_not_registered
                        }
                    }
                }
            }
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        user_profile_func, _ = find_endpoint('user/profile/')
        payment_func, _ = find_endpoint('payments/')
        admin_v1_profile_avatar_func, _ = find_endpoint('admin/v1/profile/avatar')
        admin_v1_id_func, _ = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        admin_v2_users_list_registered_func, _ = find_endpoint('admin/v1/users/list/registered/')
        admin_v2_users_detail_not_registered_func, _ = find_endpoint('admin/v1/users/detail/not-registered')

        self.assertEqual(user_id_profile_id_func, user_id_profile_id)
        self.assertEqual(user_profile_func, user_profile)
        self.assertEqual(payment_func, payment)
        self.assertEqual(admin_v1_profile_avatar_func, admin_v1_profile_avatar)
        self.assertEqual(admin_v1_id_func, admin_v1_id)
        self.assertEqual(admin_v2_users_list_registered_func, admin_v2_users_list_registered)
        self.assertEqual(admin_v2_users_detail_not_registered_func, admin_v2_users_detail_not_registered)
        config['urls'] = {}

    def test_find_endpoint_success_path(self):
        def user_id_profile_id(): pass

        def user_profile(): pass

        def payment(): pass

        def admin_v1_profile_avatar(): pass

        def admin_v1_id(): pass

        def admin_v2_users_list_registered(): pass

        def admin_v2_users_detail_not_registered(): pass

        from panther.configs import config

        config['urls'] = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': user_id_profile_id
                    }
                },
                'profile': user_profile,
                'list': ...
            },
            'payments': payment,
            'notifications': ...,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': admin_v1_profile_avatar
                    },
                    '<user_id>': admin_v1_id,
                    'users': {
                        'list': {
                            'registered': admin_v2_users_list_registered,
                            'not-registered': ...
                        },
                        'detail': {
                            'registered': ...,
                            'not-registered': admin_v2_users_detail_not_registered
                        }
                    }
                }
            }
        }
        _, user_id_profile_id_path = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        _, user_profile_path = find_endpoint('user/profile/')
        _, payment_path = find_endpoint('payments/')
        _, admin_v1_profile_avatar_path = find_endpoint('admin/v1/profile/avatar')
        _, admin_v1_id_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        _, admin_v2_users_list_registered_path = find_endpoint('admin/v1/users/list/registered/')
        _, admin_v2_users_detail_not_registered_path = find_endpoint('admin/v1/users/detail/not-registered')

        self.assertEqual(user_id_profile_id_path, f'user/<user_id>/profile/<id>/')
        self.assertEqual(user_profile_path, 'user/profile/')
        self.assertEqual(payment_path, 'payments/')
        self.assertEqual(admin_v1_profile_avatar_path, 'admin/v1/profile/avatar/')
        self.assertEqual(admin_v1_id_path, 'admin/v1/<user_id>/')
        self.assertEqual(admin_v2_users_list_registered_path, 'admin/v1/users/list/registered/')
        self.assertEqual(admin_v2_users_detail_not_registered_path, 'admin/v1/users/detail/not-registered/')
        config['urls'] = {}

    def test_find_endpoint_not_found(self):
        def temp_func(): pass

        from panther.configs import config

        config['urls'] = {
            'user': {
                'list': temp_func
            }
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        user_profile_func, _ = find_endpoint('user/profile/')
        payment_func, _ = find_endpoint('payments/')
        admin_v1_profile_avatar_func, _ = find_endpoint('admin/v1/profile/avatar')
        admin_v1_id_func, _ = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        admin_v2_users_list_registered_func, _ = find_endpoint('admin/v1/users/list/registered/')
        admin_v2_users_detail_not_registered_func, _ = find_endpoint('admin/v1/users/detail/not-registered')

        self.assertIsNone(user_id_profile_id_func)
        self.assertIsNone(user_profile_func)
        self.assertIsNone(payment_func)
        self.assertIsNone(admin_v1_profile_avatar_func)
        self.assertIsNone(admin_v1_id_func)
        self.assertIsNone(admin_v2_users_list_registered_func)
        self.assertIsNone(admin_v2_users_detail_not_registered_func)
        config['urls'] = {}

    def test_find_endpoint_not_found_path(self):
        def temp_func(): pass

        from panther.configs import config

        config['urls'] = {
            'user': {
                'list': temp_func
            }
        }
        _, user_id_profile_id_path = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        _, user_profile_path = find_endpoint('user/profile/')
        _, payment_path = find_endpoint('payments/')
        _, admin_v1_profile_avatar_path = find_endpoint('admin/v1/profile/avatar')
        _, admin_v1_id_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        _, admin_v2_users_list_registered_path = find_endpoint('admin/v1/users/list/registered/')
        _, admin_v2_users_detail_not_registered_path = find_endpoint('admin/v1/users/detail/not-registered')

        self.assertEqual(user_id_profile_id_path, '')
        self.assertEqual(user_profile_path, '')
        self.assertEqual(payment_path, '')
        self.assertEqual(admin_v1_profile_avatar_path, '')
        self.assertEqual(admin_v1_id_path, '')
        self.assertEqual(admin_v2_users_list_registered_path, '')
        self.assertEqual(admin_v2_users_detail_not_registered_path, '')
        config['urls'] = {}

    def test_collect_path_variables(self):
        def temp_func(): pass

        from panther.configs import config

        config['urls'] = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': temp_func
                    }
                },
            }
        }

        _user_id = random.randint(0, 100)
        _id = random.randint(0, 100)
        request_path = f'user/{_user_id}/profile/{_id}'

        _, found_path = find_endpoint(request_path)
        path_variables = collect_path_variables(request_path=request_path, found_path=found_path)

        self.assertIsInstance(path_variables, dict)

        self.assertTrue('user_id' in path_variables)
        self.assertTrue('id' in path_variables)

        self.assertEqual(path_variables['user_id'], str(_user_id))
        self.assertEqual(path_variables['id'], str(_id))

        config['urls'] = {}



import os
import bson
import faker
import random
from unittest import TestCase
from pydantic import field_validator, Field, BaseModel as PydanticBaseModel

from panther.configs import config
from panther.db.connection import DBSession
from panther.db.queries.mongodb_queries import BaseMongoDBQuery


f = faker.Faker()


class Model(PydanticBaseModel):
    id: str | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value):
        if isinstance(value, str):
            try:
                bson.ObjectId(value)
            except Exception:
                raise ValueError('Invalid ObjectId')
        elif not isinstance(value, bson.ObjectId):
            raise ValueError('ObjectId required')
        return str(value)

    @property
    def _id(self):
        return bson.ObjectId(self.id) if self.id else None


class Book(Model, BaseMongoDBQuery):
    name: str
    author: str
    pages_count: int


class TestPantherDB(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        config['db_engine'] = 'mongodb'
        cls.db = DBSession(db_url=f'mongodb://127.0.0.1:27017/test')

    @classmethod
    def tearDownClass(cls) -> None:
        cls.db.close()

    def test_insert_one(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        book = Book.insert_one(name=name, author=author, pages_count=pages_count)

        self.assertIsInstance(book, Book)
        self.assertIsNotNone(book.id)
        self.assertEqual(book.name, name)
        self.assertEqual(book.pages_count, pages_count)



from panther.app import API
from panther.configs import config


@API()
async def list_models():
    result = list()
    for i, m in enumerate(config['models']):
        data = dict()
        data['name'] = m['name']
        data['app'] = '.'.join(a for a in m['app'])
        data['path'] = m['path']
        data['index'] = i
        result.append(data)
    return result


@API()
async def model_list(index: int):
    model = config['models'][index]['class']
    result ={
        'fields': {k: getattr(v.annotation, '__name__', str(v.annotation)) for k, v in model.model_fields.items()},
    }
    if data := model.find():
        result['data'] = data
    else:
        result['data'] = []
    return result


@API()
async def model_retrieve(index: int, id: int):
    model = config['models'][index]['class']
    return model.find_one(id=id)


